@@ -1,29 +1,93 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 'use strict';
 
 const util = require('util');
 const net = require('net');
 const url = require('url');
-const HTTPParser = process.binding('http_parser').HTTPParser;
+const { HTTPParser } = process.binding('http_parser');
 const assert = require('assert').ok;
-const common = require('_http_common');
-const httpSocketSetup = common.httpSocketSetup;
-const parsers = common.parsers;
-const freeParser = common.freeParser;
-const debug = common.debug;
-const OutgoingMessage = require('_http_outgoing').OutgoingMessage;
+const {
+  _checkIsHttpToken: checkIsHttpToken,
+  debug,
+  freeParser,
+  httpSocketSetup,
+  parsers
+} = require('_http_common');
+const { OutgoingMessage } = require('_http_outgoing');
 const Agent = require('_http_agent');
-const Buffer = require('buffer').Buffer;
+const { Buffer } = require('buffer');
+const { urlToOptions, searchParamsSymbol } = require('internal/url');
+const { outHeadersKey, ondrain } = require('internal/http');
+const { nextTick } = require('internal/process/next_tick');
 
+// The actual list of disallowed characters in regexp form is more like:
+//    /[^A-Za-z0-9\-._~!$&'()*+,;=/:@]/
+// with an additional rule for ignoring percentage-escaped characters, but
+// that's a) hard to capture in a regular expression that performs well, and
+// b) possibly too restrictive for real-world usage. So instead we restrict the
+// filter to just control characters and spaces.
+//
+// This function is used in the case of small paths, where manual character code
+// checks can greatly outperform the equivalent regexp (tested in V8 5.4).
+function isInvalidPath(s) {
+  var i = 0;
+  if (s.charCodeAt(0) <= 32) return true;
+  if (++i >= s.length) return false;
+  if (s.charCodeAt(1) <= 32) return true;
+  if (++i >= s.length) return false;
+  if (s.charCodeAt(2) <= 32) return true;
+  if (++i >= s.length) return false;
+  if (s.charCodeAt(3) <= 32) return true;
+  if (++i >= s.length) return false;
+  if (s.charCodeAt(4) <= 32) return true;
+  if (++i >= s.length) return false;
+  if (s.charCodeAt(5) <= 32) return true;
+  ++i;
+  for (; i < s.length; ++i)
+    if (s.charCodeAt(i) <= 32) return true;
+  return false;
+}
+
+function validateHost(host, name) {
+  if (host != null && typeof host !== 'string') {
+    throw new TypeError(
+      `"options.${name}" must either be a string, undefined or null`);
+  }
+  return host;
+}
 
 function ClientRequest(options, cb) {
-  var self = this;
-  OutgoingMessage.call(self);
+  OutgoingMessage.call(this);
 
   if (typeof options === 'string') {
     options = url.parse(options);
     if (!options.hostname) {
       throw new Error('Unable to determine the domain name');
     }
+  } else if (options && options[searchParamsSymbol] &&
+             options[searchParamsSymbol][searchParamsSymbol]) {
+    // url.URL instance
+    options = urlToOptions(options);
   } else {
     options = util._extend({}, options);
   }
@@ -32,69 +96,93 @@ function ClientRequest(options, cb) {
   var defaultAgent = options._defaultAgent || Agent.globalAgent;
   if (agent === false) {
     agent = new defaultAgent.constructor();
-  } else if ((agent === null || agent === undefined) &&
-             typeof options.createConnection !== 'function') {
-    agent = defaultAgent;
+  } else if (agent === null || agent === undefined) {
+    if (typeof options.createConnection !== 'function') {
+      agent = defaultAgent;
+    }
+    // Explicitly pass through this statement as agent will not be used
+    // when createConnection is provided.
+  } else if (typeof agent.addRequest !== 'function') {
+    throw new TypeError(
+      'Agent option must be an Agent-like object, undefined, or false.'
+    );
   }
-  self.agent = agent;
+  this.agent = agent;
 
   var protocol = options.protocol || defaultAgent.protocol;
   var expectedProtocol = defaultAgent.protocol;
-  if (self.agent && self.agent.protocol)
-    expectedProtocol = self.agent.protocol;
+  if (this.agent && this.agent.protocol)
+    expectedProtocol = this.agent.protocol;
 
-  if (options.path && /[\u0000-\u0020]/.test(options.path)) {
-    // The actual regex is more like /[^A-Za-z0-9\-._~!$&'()*+,;=/:@]/
-    // with an additional rule for ignoring percentage-escaped characters
-    // but that's a) hard to capture in a regular expression that performs
-    // well, and b) possibly too restrictive for real-world usage.
-    // Restrict the filter to control characters and spaces.
-    throw new TypeError('Request path contains unescaped characters');
-  } else if (protocol !== expectedProtocol) {
+  var path;
+  if (options.path) {
+    path = String(options.path);
+    var invalidPath;
+    if (path.length <= 39) { // Determined experimentally in V8 5.4
+      invalidPath = isInvalidPath(path);
+    } else {
+      invalidPath = /[\u0000-\u0020]/.test(path);
+    }
+    if (invalidPath)
+      throw new TypeError('Request path contains unescaped characters');
+  }
+
+  if (protocol !== expectedProtocol) {
     throw new Error('Protocol "' + protocol + '" not supported. ' +
                     'Expected "' + expectedProtocol + '"');
   }
 
-  const defaultPort = options.defaultPort ||
-                      self.agent && self.agent.defaultPort;
+  var defaultPort = options.defaultPort ||
+                    this.agent && this.agent.defaultPort;
 
   var port = options.port = options.port || defaultPort || 80;
-  var host = options.host = options.hostname || options.host || 'localhost';
+  var host = options.host = validateHost(options.hostname, 'hostname') ||
+                            validateHost(options.host, 'host') || 'localhost';
 
-  if (options.setHost === undefined) {
-    var setHost = true;
+  var setHost = (options.setHost === undefined);
+
+  this.socketPath = options.socketPath;
+  this.timeout = options.timeout;
+
+  var method = options.method;
+  var methodIsString = (typeof method === 'string');
+  if (method != null && !methodIsString) {
+    throw new TypeError('Method must be a string');
   }
 
-  self.socketPath = options.socketPath;
-  self.timeout = options.timeout;
-
-  var method = self.method = (options.method || 'GET').toUpperCase();
-  if (!common._checkIsHttpToken(method)) {
-    throw new TypeError('Method must be a valid HTTP token');
+  if (methodIsString && method) {
+    if (!checkIsHttpToken(method)) {
+      throw new TypeError('Method must be a valid HTTP token');
+    }
+    method = this.method = method.toUpperCase();
+  } else {
+    method = this.method = 'GET';
   }
-  self.path = options.path || '/';
+
+  this.path = options.path || '/';
   if (cb) {
-    self.once('response', cb);
+    this.once('response', cb);
   }
 
-  if (!Array.isArray(options.headers)) {
+  var headersArray = Array.isArray(options.headers);
+  if (!headersArray) {
     if (options.headers) {
       var keys = Object.keys(options.headers);
-      for (var i = 0, l = keys.length; i < l; i++) {
+      for (var i = 0; i < keys.length; i++) {
         var key = keys[i];
-        self.setHeader(key, options.headers[key]);
+        this.setHeader(key, options.headers[key]);
       }
     }
     if (host && !this.getHeader('host') && setHost) {
       var hostHeader = host;
-      var posColon = -1;
 
       // For the Host header, ensure that IPv6 addresses are enclosed
       // in square brackets, as defined by URI formatting
       // https://tools.ietf.org/html/rfc3986#section-3.2.2
-      if (-1 !== (posColon = hostHeader.indexOf(':')) &&
-          -1 !== (posColon = hostHeader.indexOf(':', posColon)) &&
-          '[' !== hostHeader[0]) {
+      var posColon = hostHeader.indexOf(':');
+      if (posColon !== -1 &&
+          hostHeader.indexOf(':', posColon + 1) !== -1 &&
+          hostHeader.charCodeAt(0) !== 91/*'['*/) {
         hostHeader = `[${hostHeader}]`;
       }
 
@@ -106,7 +194,6 @@ function ClientRequest(options, cb) {
   }
 
   if (options.auth && !this.getHeader('Authorization')) {
-    //basic auth
     this.setHeader('Authorization', 'Basic ' +
                    Buffer.from(options.auth).toString('base64'));
   }
@@ -116,111 +203,101 @@ function ClientRequest(options, cb) {
       method === 'DELETE' ||
       method === 'OPTIONS' ||
       method === 'CONNECT') {
-    self.useChunkedEncodingByDefault = false;
+    this.useChunkedEncodingByDefault = false;
   } else {
-    self.useChunkedEncodingByDefault = true;
+    this.useChunkedEncodingByDefault = true;
   }
 
-  if (Array.isArray(options.headers)) {
-    self._storeHeader(self.method + ' ' + self.path + ' HTTP/1.1\r\n',
+  if (headersArray) {
+    this._storeHeader(this.method + ' ' + this.path + ' HTTP/1.1\r\n',
                       options.headers);
-  } else if (self.getHeader('expect')) {
-    self._storeHeader(self.method + ' ' + self.path + ' HTTP/1.1\r\n',
-                      self._renderHeaders());
+  } else if (this.getHeader('expect')) {
+    if (this._header) {
+      throw new Error('Can\'t render headers after they are sent to the ' +
+                      'client');
+    }
+
+    this._storeHeader(this.method + ' ' + this.path + ' HTTP/1.1\r\n',
+                      this[outHeadersKey]);
   }
 
+  this._ended = false;
+  this.res = null;
+  this.aborted = undefined;
+  this.timeoutCb = null;
+  this.upgradeOrConnect = false;
+  this.parser = null;
+  this.maxHeadersCount = null;
+
   var called = false;
-  if (self.socketPath) {
-    self._last = true;
-    self.shouldKeepAlive = false;
-    const optionsPath = {
-      path: self.socketPath,
-      timeout: self.timeout
-    };
-    const newSocket = self.agent.createConnection(optionsPath, oncreate);
-    if (newSocket && !called) {
-      called = true;
-      self.onSocket(newSocket);
-    } else {
+
+  var oncreate = (err, socket) => {
+    if (called)
+      return;
+    called = true;
+    if (err) {
+      process.nextTick(() => this.emit('error', err));
       return;
     }
-  } else if (self.agent) {
+    this.onSocket(socket);
+    this._deferToConnect(null, null, () => this._flush());
+  };
+
+  if (this.agent) {
     // If there is an agent we should default to Connection:keep-alive,
     // but only if the Agent will actually reuse the connection!
     // If it's not a keepAlive agent, and the maxSockets==Infinity, then
     // there's never a case where this socket will actually be reused
-    if (!self.agent.keepAlive && !Number.isFinite(self.agent.maxSockets)) {
-      self._last = true;
-      self.shouldKeepAlive = false;
+    if (!this.agent.keepAlive && !Number.isFinite(this.agent.maxSockets)) {
+      this._last = true;
+      this.shouldKeepAlive = false;
     } else {
-      self._last = false;
-      self.shouldKeepAlive = true;
+      this._last = false;
+      this.shouldKeepAlive = true;
     }
-    self.agent.addRequest(self, options);
+    this.agent.addRequest(this, options);
   } else {
     // No agent, default to Connection:close.
-    self._last = true;
-    self.shouldKeepAlive = false;
+    this._last = true;
+    this.shouldKeepAlive = false;
     if (typeof options.createConnection === 'function') {
       const newSocket = options.createConnection(options, oncreate);
       if (newSocket && !called) {
         called = true;
-        self.onSocket(newSocket);
+        this.onSocket(newSocket);
       } else {
         return;
       }
     } else {
       debug('CLIENT use net.createConnection', options);
-      self.onSocket(net.createConnection(options));
+      this.onSocket(net.createConnection(options));
     }
   }
 
-  function oncreate(err, socket) {
-    if (called)
-      return;
-    called = true;
-    if (err) {
-      process.nextTick(function() {
-        self.emit('error', err);
-      });
-      return;
-    }
-    self.onSocket(socket);
-    self._deferToConnect(null, null, function() {
-      self._flush();
-      self = null;
-    });
-  }
-
-  self._deferToConnect(null, null, function() {
-    self._flush();
-    self = null;
-  });
-
-  this._ended = false;
+  this._deferToConnect(null, null, () => this._flush());
 }
 
 util.inherits(ClientRequest, OutgoingMessage);
 
-exports.ClientRequest = ClientRequest;
 
-ClientRequest.prototype.aborted = undefined;
-
-ClientRequest.prototype._finish = function() {
+ClientRequest.prototype._finish = function _finish() {
   DTRACE_HTTP_CLIENT_REQUEST(this, this.connection);
   LTTNG_HTTP_CLIENT_REQUEST(this, this.connection);
   COUNTER_HTTP_CLIENT_REQUEST();
   OutgoingMessage.prototype._finish.call(this);
 };
 
-ClientRequest.prototype._implicitHeader = function() {
+ClientRequest.prototype._implicitHeader = function _implicitHeader() {
+  if (this._header) {
+    throw new Error('Can\'t render headers after they are sent to the client');
+  }
   this._storeHeader(this.method + ' ' + this.path + ' HTTP/1.1\r\n',
-                    this._renderHeaders());
+                    this[outHeadersKey]);
 };
 
-ClientRequest.prototype.abort = function() {
-  if (this.aborted === undefined) {
-    process.nextTick(emitAbortNT, this);
+ClientRequest.prototype.abort = function abort() {
+  if (!this.aborted) {
+    process.nextTick(emitAbortNT.bind(this));
   }
   // Mark as aborting so we can avoid sending queued request data
   // This is used as a truthy flag elsewhere. The use of Date.now is for
@@ -228,12 +305,13 @@ ClientRequest.prototype.abort = function() {
   this.aborted = Date.now();
 
   // If we're aborting, we don't care about any more response data.
-  if (this.res)
+  if (this.res) {
     this.res._dump();
-  else
+  } else {
     this.once('response', function(res) {
       res._dump();
     });
+  }
 
   // In the event that we don't have a socket, we will pop out of
   // the request queue through handling in onSocket.
@@ -244,8 +322,8 @@ ClientRequest.prototype.abort = function() {
 };
 
 
-function emitAbortNT(self) {
-  self.emit('abort');
+function emitAbortNT() {
+  this.emit('abort');
 }
 
 
@@ -272,7 +350,7 @@ function socketCloseListener() {
   req.emit('close');
   if (req.res && req.res.readable) {
     // Socket closed before we emitted 'end' below.
-    req.res.emit('aborted');
+    if (!req.res.complete) req.res.emit('aborted');
     var res = req.res;
     res.on('end', function() {
       res.emit('close');
@@ -361,7 +439,7 @@ function socketOnData(d) {
 
   var ret = parser.execute(d);
   if (ret instanceof Error) {
-    debug('parse error');
+    debug('parse error', ret);
     freeParser(parser, req, socket);
     socket.destroy();
     req.emit('error', ret);
@@ -374,7 +452,9 @@ function socketOnData(d) {
 
     socket.removeListener('data', socketOnData);
     socket.removeListener('end', socketOnEnd);
+    socket.removeListener('drain', ondrain);
     parser.finish();
+    freeParser(parser, req, socket);
 
     var bodyHead = d.slice(bytesParsed, d.length);
 
@@ -390,6 +470,7 @@ function socketOnData(d) {
       // TODO(isaacs): Need a way to reset a stream to fresh state
       // IE, not flowing, and not explicitly paused.
       socket._readableState.flowing = null;
+      socket._httpMessage = null;
 
       req.emit(eventName, res, socket, bodyHead);
       req.emit('close');
@@ -397,7 +478,6 @@ function socketOnData(d) {
       // Got Upgrade header or CONNECT method, but have no handler.
       socket.destroy();
     }
-    freeParser(parser, req, socket);
   } else if (parser.incoming && parser.incoming.complete &&
              // When the status code is 100 (Continue), the server will
              // send a final response after this client sends a request
@@ -415,7 +495,6 @@ function parserOnIncomingClient(res, shouldKeepAlive) {
   var socket = this.socket;
   var req = socket._httpMessage;
 
-
   // propagate "domain" setting...
   if (req.domain && !res.domain) {
     debug('setting "res.domain"');
@@ -428,29 +507,22 @@ function parserOnIncomingClient(res, shouldKeepAlive) {
     // We already have a response object, this means the server
     // sent a double response.
     socket.destroy();
-    return;
+    return 0;  // No special treatment.
   }
   req.res = res;
 
   // Responses to CONNECT request is handled as Upgrade.
-  if (req.method === 'CONNECT') {
+  const method = req.method;
+  if (method === 'CONNECT') {
     res.upgrade = true;
-    return 2; // skip body, and the rest
+    return 2;  // Skip body and treat as Upgrade.
   }
 
-  // Responses to HEAD requests are crazy.
-  // HEAD responses aren't allowed to have an entity-body
-  // but *can* have a content-length which actually corresponds
-  // to the content-length of the entity-body had the request
-  // been a GET.
-  var isHeadResponse = req.method === 'HEAD';
-  debug('AGENT isHeadResponse', isHeadResponse);
-
   if (res.statusCode === 100) {
     // restart the parser, as this is a continue message.
-    delete req.res; // Clear res so that we don't hit double-responses.
+    req.res = null; // Clear res so that we don't hit double-responses.
     req.emit('continue');
-    return true;
+    return 1;  // Skip body but don't treat as Upgrade.
   }
 
   if (req.shouldKeepAlive && !shouldKeepAlive && !req.upgradeOrConnect) {
@@ -460,7 +532,6 @@ function parserOnIncomingClient(res, shouldKeepAlive) {
     req.shouldKeepAlive = false;
   }
 
-
   DTRACE_HTTP_CLIENT_RESPONSE(socket, req);
   LTTNG_HTTP_CLIENT_RESPONSE(socket, req);
   COUNTER_HTTP_CLIENT_RESPONSE();
@@ -478,7 +549,10 @@ function parserOnIncomingClient(res, shouldKeepAlive) {
   if (!handled)
     res._dump();
 
-  return isHeadResponse;
+  if (method === 'HEAD')
+    return 1;  // Skip body but don't treat as Upgrade.
+
+  return 0;  // No special treatment.
 }
 
 // client
@@ -488,7 +562,10 @@ function responseKeepAlive(res, req) {
   if (!req.shouldKeepAlive) {
     if (socket.writable) {
       debug('AGENT socket.destroySoon()');
-      socket.destroySoon();
+      if (typeof socket.destroySoon === 'function')
+        socket.destroySoon();
+      else
+        socket.end();
     }
     assert(!socket.writable);
   } else {
@@ -500,9 +577,12 @@ function responseKeepAlive(res, req) {
     socket.removeListener('close', socketCloseListener);
     socket.removeListener('error', socketErrorListener);
     socket.once('error', freeSocketErrorListener);
+    // There are cases where _handle === null. Avoid those. Passing null to
+    // nextTick() will call getDefaultTriggerAsyncId() to retrieve the id.
+    const asyncId = socket._handle ? socket._handle.getAsyncId() : null;
     // Mark this socket as available, AFTER user-added end
     // handlers have a chance to run.
-    process.nextTick(emitFreeNT, socket);
+    nextTick(asyncId, emitFreeNT, socket);
   }
 }
 
@@ -573,14 +653,14 @@ function tickOnSocket(req, socket) {
   req.emit('socket', socket);
 }
 
-ClientRequest.prototype.onSocket = function(socket) {
+ClientRequest.prototype.onSocket = function onSocket(socket) {
   process.nextTick(onSocketNT, this, socket);
 };
 
 function onSocketNT(req, socket) {
   if (req.aborted) {
     // If we were aborted while waiting for a socket, skip the whole thing.
-    if (req.socketPath || !req.agent) {
+    if (!req.agent) {
       socket.destroy();
     } else {
       socket.emit('free');
@@ -590,44 +670,41 @@ function onSocketNT(req, socket) {
   }
 }
 
-ClientRequest.prototype._deferToConnect = function(method, arguments_, cb) {
+ClientRequest.prototype._deferToConnect = _deferToConnect;
+function _deferToConnect(method, arguments_, cb) {
   // This function is for calls that need to happen once the socket is
   // connected and writable. It's an important promisy thing for all the socket
   // calls that happen either now (when a socket is assigned) or
   // in the future (when a socket gets assigned out of the pool and is
   // eventually writable).
-  var self = this;
 
-  function callSocketMethod() {
+  const callSocketMethod = () => {
     if (method)
-      self.socket[method].apply(self.socket, arguments_);
+      this.socket[method].apply(this.socket, arguments_);
 
     if (typeof cb === 'function')
       cb();
-  }
+  };
 
-  var onSocket = function() {
-    if (self.socket.writable) {
+  const onSocket = () => {
+    if (this.socket.writable) {
       callSocketMethod();
     } else {
-      self.socket.once('connect', callSocketMethod);
+      this.socket.once('connect', callSocketMethod);
     }
   };
 
-  if (!self.socket) {
-    self.once('socket', onSocket);
+  if (!this.socket) {
+    this.once('socket', onSocket);
   } else {
     onSocket();
   }
-};
+}
 
-ClientRequest.prototype.setTimeout = function(msecs, callback) {
+ClientRequest.prototype.setTimeout = function setTimeout(msecs, callback) {
   if (callback) this.once('timeout', callback);
 
-  var self = this;
-  function emitTimeout() {
-    self.emit('timeout');
-  }
+  const emitTimeout = () => this.emit('timeout');
 
   if (this.socket && this.socket.writable) {
     if (this.timeoutCb)
@@ -654,21 +731,19 @@ ClientRequest.prototype.setTimeout = function(msecs, callback) {
   return this;
 };
 
-ClientRequest.prototype.setNoDelay = function() {
-  const argsLen = arguments.length;
-  const args = new Array(argsLen);
-  for (var i = 0; i < argsLen; i++)
-    args[i] = arguments[i];
-  this._deferToConnect('setNoDelay', args);
-};
-ClientRequest.prototype.setSocketKeepAlive = function() {
-  const argsLen = arguments.length;
-  const args = new Array(argsLen);
-  for (var i = 0; i < argsLen; i++)
-    args[i] = arguments[i];
-  this._deferToConnect('setKeepAlive', args);
+ClientRequest.prototype.setNoDelay = function setNoDelay(noDelay) {
+  this._deferToConnect('setNoDelay', [noDelay]);
 };
 
-ClientRequest.prototype.clearTimeout = function(cb) {
+ClientRequest.prototype.setSocketKeepAlive =
+    function setSocketKeepAlive(enable, initialDelay) {
+      this._deferToConnect('setKeepAlive', [enable, initialDelay]);
+    };
+
+ClientRequest.prototype.clearTimeout = function clearTimeout(cb) {
   this.setTimeout(0, cb);
 };
+
+module.exports = {
+  ClientRequest
+};
