@@ -1,4 +1,7 @@
-/* eslint-disable strict */
+'use strict';
+
+const vm = require('vm');
+
 const scriptFiles = [
   'internal/v8_prof_polyfill',
   'v8/tools/splaytree',
@@ -30,9 +33,9 @@ if (process.platform === 'darwin') {
   tickArguments.push('--windows');
 }
 tickArguments.push.apply(tickArguments, process.argv.slice(1));
-script = `(function() {
+script = `(function(module, require) {
   arguments = ${JSON.stringify(tickArguments)};
   function write (s) { process.stdout.write(s) }
   ${script}
-})()`;
-eval(script);
+})`;
+vm.runInThisContext(script)(module, require);
