@@ -1,9 +1,9 @@
 'use strict';
 
-const common = require('../common');
+require('../common');
 const assert = require('assert');
 const fs = require('fs');
-const path = require('path');
+const fixtures = require('../common/fixtures');
 
 let openCount = 0;
 const _fsopen = fs.open;
@@ -11,7 +11,7 @@ const _fsclose = fs.close;
 
 const loopCount = 50;
 const totalCheck = 50;
-const emptyTxt = path.join(common.fixturesDir, 'empty.txt');
+const emptyTxt = fixtures.path('empty.txt');
 
 fs.open = function() {
   openCount++;
@@ -24,7 +24,7 @@ fs.close = function() {
 };
 
 function testLeak(endFn, callback) {
-  console.log('testing for leaks from fs.createReadStream().%s()...', endFn);
+  console.log(`testing for leaks from fs.createReadStream().${endFn}()...`);
 
   let i = 0;
   let check = 0;
