@@ -1,29 +1,27 @@
 'use strict';
 const common = require('../common');
-if (!common.hasCrypto) {
+if (!common.hasCrypto)
   common.skip('missing crypto');
-  return;
-}
 
-if (!common.hasIPv6) {
+if (!common.hasIPv6)
   common.skip('no IPv6 support');
-  return;
-}
 
 const assert = require('assert');
+const fixtures = require('../common/fixtures');
 const tls = require('tls');
 const dns = require('dns');
 
 function runTest() {
-  const ciphers = 'AECDH-NULL-SHA';
-  tls.createServer({ ciphers }, common.mustCall(function() {
+  tls.createServer({
+    cert: fixtures.readKey('agent1-cert.pem'),
+    key: fixtures.readKey('agent1-key.pem'),
+  }, common.mustCall(function() {
     this.close();
   })).listen(0, '::1', common.mustCall(function() {
     const options = {
       host: 'localhost',
       port: this.address().port,
       family: 6,
-      ciphers: ciphers,
       rejectUnauthorized: false,
     };
     // Will fail with ECONNREFUSED if the address family is not honored.
@@ -34,12 +32,11 @@ function runTest() {
   }));
 }
 
-dns.lookup('localhost', {family: 6, all: true}, (err, addresses) => {
+dns.lookup('localhost', { family: 6, all: true }, (err, addresses) => {
   if (err) {
-    if (err.code === 'ENOTFOUND') {
+    if (err.code === 'ENOTFOUND' || err.code === 'EAI_AGAIN')
       common.skip('localhost does not resolve to ::1');
-      return;
-    }
+
     throw err;
   }
 
