@@ -25,7 +25,6 @@ const common = require('./common.js');
 const fs = require('fs');
 const marked = require('marked');
 const path = require('path');
-const preprocess = require('./preprocess.js');
 const typeParser = require('./type-parser.js');
 
 module.exports = toHTML;
@@ -33,7 +32,7 @@ module.exports = toHTML;
 const STABILITY_TEXT_REG_EXP = /(.*:)\s*(\d)([\s\S]*)/;
 const DOC_CREATED_REG_EXP = /<!--\s*introduced_in\s*=\s*v([0-9]+)\.([0-9]+)\.([0-9]+)\s*-->/;
 
-// customized heading without id attribute
+// Customized heading without id attribute.
 const renderer = new marked.Renderer();
 renderer.heading = function(text, level) {
   return `<h${level}>${text}</h${level}>\n`;
@@ -42,76 +41,36 @@ marked.setOptions({
   renderer: renderer
 });
 
-// TODO(chrisdickinson): never stop vomitting / fix this.
-const gtocPath = path.resolve(path.join(
-  __dirname,
-  '..',
-  '..',
-  'doc',
-  'api',
-  '_toc.md'
-));
-var gtocLoading = null;
-var gtocData = null;
+const docPath = path.resolve(__dirname, '..', '..', 'doc');
+
+const gtocPath = path.join(docPath, 'api', '_toc.md');
+const gtocMD = fs.readFileSync(gtocPath, 'utf8').replace(/^@\/\/.*$/gm, '');
+const gtocHTML = marked(gtocMD).replace(
+  /<a href="(.*?)"/g,
+  (all, href) => `<a class="nav-${toID(href)}" href="${href}"`
+);
+
+const templatePath = path.join(docPath, 'template.html');
+const template = fs.readFileSync(templatePath, 'utf8');
+
 var docCreated = null;
 var nodeVersion = null;
 
 /**
- * opts: input, filename, template, nodeVersion.
+ * opts: input, filename, nodeVersion.
  */
 function toHTML(opts, cb) {
-  const template = opts.template;
-
   nodeVersion = opts.nodeVersion || process.version;
   docCreated = opts.input.match(DOC_CREATED_REG_EXP);
 
-  if (gtocData) {
-    return onGtocLoaded();
-  }
-
-  if (gtocLoading === null) {
-    gtocLoading = [onGtocLoaded];
-    return loadGtoc(function(err, data) {
-      if (err) throw err;
-      gtocData = data;
-      gtocLoading.forEach(function(xs) {
-        xs();
-      });
-    });
-  }
-
-  if (gtocLoading) {
-    return gtocLoading.push(onGtocLoaded);
-  }
-
-  function onGtocLoaded() {
-    const lexed = marked.lexer(opts.input);
-    fs.readFile(template, 'utf8', function(er, template) {
-      if (er) return cb(er);
-      render({
-        lexed: lexed,
-        filename: opts.filename,
-        template: template,
-        nodeVersion: nodeVersion,
-        analytics: opts.analytics,
-      }, cb);
-    });
-  }
-}
-
-function loadGtoc(cb) {
-  fs.readFile(gtocPath, 'utf8', function(err, data) {
-    if (err) return cb(err);
-
-    preprocess(gtocPath, data, function(err, data) {
-      if (err) return cb(err);
-
-      data = marked(data).replace(/<a href="(.*?)"/gm, function(a, m) {
-        return `<a class="nav-${toID(m)}" href="${m}"`;
-      });
-      return cb(null, data);
-    });
-  });
+  const lexed = marked.lexer(opts.input);
+  render({
+    lexed: lexed,
+    filename: opts.filename,
+    template: template,
+    nodeVersion: nodeVersion,
+    analytics: opts.analytics,
+  }, cb);
 }
 
 function toID(filename) {
@@ -128,7 +87,7 @@ function render(opts, cb) {
   var { lexed, filename, template } = opts;
   const nodeVersion = opts.nodeVersion || process.version;
 
-  // get the section
+  // Get the section.
   const section = getSection(lexed);
 
   filename = path.basename(filename, '.md');
@@ -136,8 +95,8 @@ function render(opts, cb) {
   parseText(lexed);
   lexed = parseLists(lexed);
 
-  // generate the table of contents.
-  // this mutates the lexed contents in-place.
+  // Generate the table of contents.
+  // This mutates the lexed contents in-place.
   buildToc(lexed, filename, function(er, toc) {
     if (er) return cb(er);
 
@@ -150,7 +109,7 @@ function render(opts, cb) {
     template = template.replace(/__TOC__/g, toc);
     template = template.replace(
       /__GTOC__/g,
-      gtocData.replace(`class="nav-${id}`, `class="nav-${id} active`)
+      gtocHTML.replace(`class="nav-${id}`, `class="nav-${id} active`)
     );
 
     if (opts.analytics) {
@@ -162,8 +121,8 @@ function render(opts, cb) {
 
     template = template.replace(/__ALTDOCS__/, altDocs(filename));
 
-    // content has to be the last thing we do with
-    // the lexed tokens, because it's destructive.
+    // Content has to be the last thing we do with the lexed tokens,
+    // because it's destructive.
     const content = marked.parser(lexed);
     template = template.replace(/__CONTENT__/g, content);
 
@@ -188,7 +147,7 @@ function analyticsScript(analytics) {
   `;
 }
 
-// replace placeholders in text tokens
+// Replace placeholders in text tokens.
 function replaceInText(text) {
   return linkJsTypeDocs(linkManPages(text));
 }
@@ -244,8 +203,8 @@ function altDocs(filename) {
   `;
 }
 
-// handle general body-text replacements
-// for example, link man page references to the actual page
+// Handle general body-text replacements.
+// For example, link man page references to the actual page.
 function parseText(lexed) {
   lexed.forEach(function(tok) {
     if (tok.type === 'table') {
@@ -272,8 +231,8 @@ function parseText(lexed) {
   });
 }
 
-// just update the list item text in-place.
-// lists that come right after a heading are what we're after.
+// Just update the list item text in-place.
+// Lists that come right after a heading are what we're after.
 function parseLists(input) {
   var state = null;
   const savedState = [];
@@ -299,8 +258,8 @@ function parseLists(input) {
         const stabilityMatch = tok.text.match(STABILITY_TEXT_REG_EXP);
         const stability = Number(stabilityMatch[2]);
         const isStabilityIndex =
-          index - 2 === headingIndex || // general
-          index - 3 === headingIndex;   // with api_metadata block
+          index - 2 === headingIndex || // General.
+          index - 3 === headingIndex;   // With api_metadata block.
 
         if (heading && isStabilityIndex) {
           heading.stability = stability;
@@ -408,23 +367,23 @@ function parseYAML(text) {
   return html.join('\n');
 }
 
-// Syscalls which appear in the docs, but which only exist in BSD / OSX
+// Syscalls which appear in the docs, but which only exist in BSD / macOS.
 const BSD_ONLY_SYSCALLS = new Set(['lchmod']);
 
-// Handle references to man pages, eg "open(2)" or "lchmod(2)"
-// Returns modified text, with such refs replace with HTML links, for example
-// '<a href="http://man7.org/linux/man-pages/man2/open.2.html">open(2)</a>'
+// Handle references to man pages, eg "open(2)" or "lchmod(2)".
+// Returns modified text, with such refs replaced with HTML links, for example
+// '<a href="http://man7.org/linux/man-pages/man2/open.2.html">open(2)</a>'.
 function linkManPages(text) {
   return text.replace(
-    /\b([a-z.]+)\((\d)([a-z]?)\)/gm,
-    (match, name, number, optionalCharacter) => {
-      // name consists of lowercase letters, number is a single digit
+    /(^|\s)([a-z.]+)\((\d)([a-z]?)\)/gm,
+    (match, beginning, name, number, optionalCharacter) => {
+      // Name consists of lowercase letters, number is a single digit.
       const displayAs = `${name}(${number}${optionalCharacter})`;
       if (BSD_ONLY_SYSCALLS.has(name)) {
-        return ` <a href="https://www.freebsd.org/cgi/man.cgi?query=${name}` +
+        return `${beginning}<a href="https://www.freebsd.org/cgi/man.cgi?query=${name}` +
           `&sektion=${number}">${displayAs}</a>`;
       } else {
-        return ` <a href="http://man7.org/linux/man-pages/man${number}` +
+        return `${beginning}<a href="http://man7.org/linux/man-pages/man${number}` +
           `/${name}.${number}${optionalCharacter}.html">${displayAs}</a>`;
       }
     });
@@ -436,7 +395,7 @@ function linkJsTypeDocs(text) {
   var typeMatches;
 
   // Handle types, for example the source Markdown might say
-  // "This argument should be a {Number} or {String}"
+  // "This argument should be a {Number} or {String}".
   for (i = 0; i < parts.length; i += 2) {
     typeMatches = parts[i].match(/\{([^}]+)\}/g);
     if (typeMatches) {
@@ -446,7 +405,7 @@ function linkJsTypeDocs(text) {
     }
   }
 
-  //XXX maybe put more stuff here?
+  // TODO: maybe put more stuff here?
   return parts.join('`');
 }
 
@@ -461,7 +420,7 @@ function parseAPIHeader(text) {
   return text;
 }
 
-// section is just the first heading
+// Section is just the first heading.
 function getSection(lexed) {
   for (var i = 0, l = lexed.length; i < l; i++) {
     var tok = lexed[i];
@@ -533,7 +492,7 @@ const numberRe = /^(\d*)/;
 function versionSort(a, b) {
   a = a.trim();
   b = b.trim();
-  let i = 0;  // common prefix length
+  let i = 0; // Common prefix length.
   while (i < a.length && i < b.length && a[i] === b[i]) i++;
   a = a.substr(i);
   b = b.substr(i);
